-- /packages/intranet-estimate-to-complete/sql/postgresql/intranet-estimate-to-complete-create.sql
--
-- Copyright (c) 2003-2008 ]project-open[
--
-- All rights reserved. Please check
-- https://www.project-open.com/license/ for details.
--
-- @author frank.bergmann@project-open.com



-----------------------------------------------------------
-- Estimate To Complete

SELECT acs_object_type__create_type (
	'im_estimate_to_complete',		-- object_type
	'Estimate to Complete',			-- pretty_name
	'Estimate to Complete',			-- pretty_plural
	'acs_object',				-- supertype
	'im_estimate_to_completes',		-- table_name
	'etc_id',				-- id_column
	'intranet-estimate-to-complete',	-- package_name
	'f',					-- abstract_p
	null,					-- type_extension_table
	'im_estimate_to_complete__name'		-- name_method
);

insert into acs_object_type_tables (object_type,table_name,id_column)
values ('im_estimate_to_complete', 'im_estimate_to_completes', 'etc_id');


insert into im_biz_object_urls (object_type, url_type, url) values (
'im_estimate_to_complete','view','/intranet-timesheet2/etc/new?form_mode=display&etc_id=');
insert into im_biz_object_urls (object_type, url_type, url) values (
'im_estimate_to_complete','edit','/intranet-timesheet2/etc/new?etc_id=');

-- Define where the system can find the status and type of the etc.
-- Allows for automatic workflow updates and display
update acs_object_types set
		status_type_table = 'im_estimate_to_completes',
		status_column = 'etc_status_id',
		type_column = 'etc_type_id',
		type_category_type = 'Intranet Estimate to Complete Type'
where object_type = 'im_estimate_to_complete';


create table im_estimate_to_completes (
	etc_id				integer
					constraint im_estimate_to_complete_id_pk
					primary key
					constraint im_estimate_to_complete_id_fk
					references acs_objects,
	etc_status_id			integer
					constraint im_estimate_to_complete_status_nn
					not null
					constraint im_estimate_to_complete_status_fk
					references im_categories,
	etc_type_id			integer
					constraint im_estimate_to_complete_type_nn
					not null
					constraint im_estimate_to_complete_type_fk
					references im_categories,
	etc_user_id			integer
					constraint im_estimate_to_complete_user_nn
					not null
					constraint im_estimate_to_complete_user_fk
					references persons,
	etc_project_id			integer
					constraint im_estimate_to_complete_project_nn
					not null
					constraint im_estimate_to_complete_project_fk
					references im_projects,
	etc_units			numeric,
	etc_uom_id			integer
					constraint im_estimate_to_complete_uom_fk
					references im_categories,

					-- project values at the time of creation
	task_planned_hours_audit	numeric,
	task_logged_hours_audit		numeric,
	task_completed_percent_audit	numeric,
	task_assigned_percent_audit	numeric,
	user_logged_hours_audit		numeric,
	user_assigned_percent_audit	numeric
);



-- Avoid duplicate entry of data
create unique index im_estimate_to_completes_un on im_estimate_to_completes (
	coalesce(etc_status_id, 0),
	coalesce(etc_type_id, 0),
	coalesce(etc_uom_id, 0),
	etc_user_id, 
	etc_project_id
);



-----------------------------------------------------------
-- Create, Drop and Name Plpg/SQL functions
--
-- These functions represent crator/destructor
-- functions for the OpenACS object system.


create or replace function im_estimate_to_complete__name(integer)
returns varchar as $$
DECLARE
	p_etc_id		alias for $1;
	v_name			varchar;
BEGIN
	select	'etc('||p.project_name||','||im_name_from_user_id(etc_user_id)||')' into v_name 
	from	im_estimate_to_completes etc,
		im_projects p
	where	etc.etc_id = p_etc_id and
		p.project_id = etc.etc_project_id;

	return v_name;
end;$$ language 'plpgsql';


create or replace function im_estimate_to_complete__new (
	integer, varchar, timestamptz, integer, varchar, integer,
	integer, integer, 
	integer, integer, numeric, integer
) returns integer as $$
DECLARE
	p_etc_id		alias for $1;		-- etc_id default null
	p_object_type		alias for $2;		-- object_type default im_estimate_to_complete
	p_creation_date 	alias for $3;		-- creation_date default now()
	p_creation_user 	alias for $4;		-- creation_user default null
	p_creation_ip		alias for $5;		-- creation_ip default null
	p_context_id		alias for $6;		-- context_id default null

	p_etc_status_id		alias for $7;
	p_etc_type_id		alias for $8;		
	p_etc_user_id		alias for $9;
	p_etc_project_id	alias for $10;
	p_etc_units		alias for $11;
	p_etc_uom_id		alias for $12;

	v_task_planned_hours_audit	numeric;
	v_task_logged_hours_audit	numeric;
	v_task_completed_percent_audit	numeric;
	v_task_assigned_percent_audit	numeric;
	v_user_logged_hours_audit	numeric;
	v_user_assigned_percent_audit	numeric;

	v_etc_id		integer;
BEGIN
	v_etc_id := acs_object__new (p_etc_id, p_object_type, p_creation_date, p_creation_user, p_creation_ip, p_context_id);

	select	 planned_units, reported_hours_cache, percent_completed,
		(select	greatest(sum(percentage), 0.01)
		from	acs_rels r,
			im_biz_object_members bom
		where	r.rel_id = bom.rel_id and
			r.object_id_one = p_etc_project_id
		),
		(	select	round(sum(h.hours),1)
			from	im_hours h
			where	h.project_id = p_etc_project_id and
				h.user_id = p_etc_user_id
		),
		(select	sum(percentage)
		from	acs_rels r,
			im_biz_object_members bom
		where	r.rel_id = bom.rel_id and
			r.object_id_two = p_etc_user_id and
			r.object_id_one = p_etc_project_id
		)
	into	 v_task_planned_hours_audit, v_task_logged_hours_audit, 
		 v_task_completed_percent_audit, v_task_assigned_percent_audit,
		 v_user_logged_hours_audit, v_user_assigned_percent_audit
	from	 im_projects p
		 LEFT OUTER JOIN im_timesheet_tasks t ON (p.project_id = t.task_id)
	where	 p.project_id = p_etc_project_id;

	insert into im_estimate_to_completes (
		etc_id, etc_status_id, etc_type_id, etc_user_id, etc_project_id, etc_units, etc_uom_id,
		task_planned_hours_audit, task_logged_hours_audit, 
		task_completed_percent_audit, task_assigned_percent_audit,
		user_logged_hours_audit, user_assigned_percent_audit
	) values (
		v_etc_id, p_etc_status_id, p_etc_type_id, p_etc_user_id, p_etc_project_id, p_etc_units, p_etc_uom_id,
		v_task_planned_hours_audit, v_task_logged_hours_audit, 
		v_task_completed_percent_audit, v_task_assigned_percent_audit,
		v_user_logged_hours_audit, v_user_assigned_percent_audit
	);

	return v_etc_id;
END;$$ language 'plpgsql';



create or replace function im_estimate_to_complete_new (
	integer, integer, numeric
) returns integer as $$
DECLARE
	p_user_id		alias for $1;
	p_project_id		alias for $2;
	p_units			alias for $3;

	v_etc_id		integer;
BEGIN
	v_etc_id := im_estimate_to_complete__new (
		null, 'im_estimate_to_complete', now(), p_user_id, null, null,
		44000, 44100, p_user_id, p_project_id, p_units, 320
	);
	return v_etc_id;
END;$$ language 'plpgsql';




create or replace function im_estimate_to_complete__delete(integer)
returns integer as $$
DECLARE
	p_etc_id	alias for $1;
BEGIN
	-- Delete any data related to the object
	delete from im_estimate_to_completes
	where	etc_id = p_etc_id;

	-- Finally delete the object iself
	PERFORM acs_object__delete(p_etc_id);

	return 0;
end;$$ language 'plpgsql';



-----------------------------------------------------------
-- Categories
--

-- 44000-44999  Intranet Estimate to Complete (1000)
-- 44000-44099  Intranet Estimate to Complete Status (100)
-- 44100-44199  Intranet Estimate to Complete Type (100)
-- 44200-44099  free (800)

SELECT im_category_new(44000, 'ETC Active', 'Intranet Estimate to Complete Status');

SELECT im_category_new(44100, 'ETC Timesheet', 'Intranet Estimate to Complete Type');


-----------------------------------------------------------
-- 
-----------------------------------------------------------


select im_component_plugin__new (
	null, 'im_component_plugin', now(), null, null, null,
	'Estimate to Complete Details',		-- plugin_name
	'intranet-estimate-to-complete',	-- package_name
	'left',					-- location
	'/intranet-timesheet2-tasks/new',	-- page_url
	null,					-- view_name
	100,					-- sort_order
	'im_estimate_to_complete_task_details_component $task_id'
);

select acs_permission__grant_permission(
	(select plugin_id from im_component_plugins where plugin_name = 'Estimate to Complete Details'),
	(select group_id from groups where group_name = 'Employees'),
	'read'
);


select im_component_plugin__new (
	null, 'im_component_plugin', now(), null, null, null,
	'Estimate to Complete History',		-- plugin_name
	'intranet-estimate-to-complete',	-- package_name
	'left',					-- location
	'/intranet-timesheet2-tasks/new',	-- page_url
	null,					-- view_name
	110,					-- sort_order
	'im_estimate_to_complete_task_history_component $task_id'
);

select acs_permission__grant_permission(
	(select plugin_id from im_component_plugins where plugin_name = 'Estimate to Complete History'),
	(select group_id from groups where group_name = 'Employees'),
	'read'
);


-----------------------------------------------------------
-- 
-----------------------------------------------------------

-- Calculate the ETC for a task or project
-- based on the entries of the user.
-- Planned ETC is used as a default if there
-- are not manual entries.
create or replace function im_estimate_to_complete__user_etc(integer)
returns numeric as $$
DECLARE
	p_project_id			alias for $1;

	v_etc				numeric;
	v_etc_user			numeric;
	v_assigned_percentage_task	numeric;
	v_planned_hours_task		numeric;
	v_reported_hours		numeric;
	v_percent_completed		numeric;
	v_sub_project_count		integer;
	row				record;
BEGIN
	-- Get some values from the task
	select	p.reported_hours_cache,
		(	select	sum(coalesce(planned_units, 0.0))
			from	im_projects pp, im_timesheet_tasks pt
			where	pp.project_id = pt.task_id and
				pp.tree_sortkey between p.tree_sortkey and tree_right(p.tree_sortkey) and
				not exists (select project_id from im_projects ppp where ppp.parent_id = pp.project_id)
		),
		p.percent_completed,
		(select count(*) from im_projects sub_p where sub_p.parent_id = p.project_id)
	into	v_reported_hours, v_planned_hours_task, v_percent_completed, v_sub_project_count
	from	im_projects p
	where	project_id = p_project_id;
--	RAISE NOTICE 'im_estimate_to_complete__user_etc: project_id=%, project_name=%', p_project_id, acs_object__name(p_project_id);
--	RAISE NOTICE 'im_estimate_to_complete__user_etc: planned_hours_task=%', v_planned_hours_task;
--	RAISE NOTICE 'im_estimate_to_complete__user_etc: reported_hours=%', v_reported_hours;
--	RAISE NOTICE 'im_estimate_to_complete__user_etc: percent_completed=%', v_percent_completed;

	-- Is this a super-task? Then descend recursively
	v_etc := 0.0;
	IF v_sub_project_count > 0 THEN 
		FOR row IN
		    	select	pp.project_id
			from	im_projects pp
			where	pp.parent_id = p_project_id
		LOOP
			v_etc := v_etc + coalesce(im_estimate_to_complete__user_etc(row.project_id), 0.0);
		END LOOP;
		return v_etc;
	END IF;

	-- Calculate the sum of all assignments to the project
	select	sum(bom.percentage) into v_assigned_percentage_task
	from	acs_rels r,
		im_biz_object_members bom
	where	r.rel_id = bom.rel_id and
		r.object_id_one = p_project_id and
		bom.percentage is not null;
--	RAISE NOTICE 'im_estimate_to_complete__user_etc: assigned_percentage_task=%', v_assigned_percentage_task;
--	RAISE NOTICE 'im_estimate_to_complete__user_etc: =%', v_;

	IF v_assigned_percentage_task = 0 THEN v_assigned_percentage_task := 0.01; END IF;
	FOR row IN
		select	pe.person_id,
			bom.percentage,
			v_planned_hours_task * (100 - v_percent_completed) * bom.percentage * 0.01 / v_assigned_percentage_task as planned_etc_user,
			(select etc_units from im_estimate_to_completes te where te.etc_project_id = p.project_id and te.etc_user_id = pe.person_id) as logged_etc_user
		from	acs_rels r,
			im_biz_object_members bom,
			persons pe,
			im_projects p
		where	p.project_id = p_project_id and
			r.rel_id = bom.rel_id and
			r.object_id_one = p.project_id and
			r.object_id_two = pe.person_id and
			bom.percentage is not null
	LOOP
--		RAISE NOTICE 'im_estimate_to_complete__user_etc: %: ass=%, planned_etc=%, logged_etc=%', 
--			im_name_from_user_id(row.person_id), row.percentage, row.planned_etc_user, row.logged_etc_user;
		v_etc_user := row.planned_etc_user;
		IF row.logged_etc_user is not null THEN v_etc_user := row.logged_etc_user; END IF;

		v_etc := v_etc + v_etc_user;
	END LOOP;

	return v_etc;
end;$$ language 'plpgsql';

-- Brainstorm concepts
-- select im_estimate_to_complete__user_etc(37483);

select im_estimate_to_complete__user_etc(37459);








-- Calculate the ETC for a user and task or project based on the entries of the user.
-- Planned ETC is used as a default if there are not manual entries.
create or replace function im_estimate_to_complete__project_user_etc(integer, integer)
returns numeric as $$
DECLARE
	p_project_id			alias for $1;
	p_user_id			alias for $2;

	v_etc				numeric;
	v_etc_user			numeric;
	v_assigned_percentage_task	numeric;
	v_planned_hours_task		numeric;
	v_reported_hours		numeric;
	v_percent_completed		numeric;
	v_sub_project_count		integer;
	row				record;
BEGIN
	-- Get some values from the task
	select	p.reported_hours_cache,
		(	select	sum(coalesce(planned_units, 0.0))
			from	im_projects pp, im_timesheet_tasks pt
			where	pp.project_id = pt.task_id and
				pp.tree_sortkey between p.tree_sortkey and tree_right(p.tree_sortkey) and
				not exists (select project_id from im_projects ppp where ppp.parent_id = pp.project_id)
		),
		p.percent_completed,
		(select count(*) from im_projects sub_p where sub_p.parent_id = p.project_id)
	into	v_reported_hours, v_planned_hours_task, v_percent_completed, v_sub_project_count
	from	im_projects p
	where	project_id = p_project_id;

	-- Is this a super-task? Then descend recursively
	v_etc := 0.0;
	IF v_sub_project_count > 0 THEN 
		FOR row IN
		    	select	pp.project_id
			from	im_projects pp
			where	pp.parent_id = p_project_id
		LOOP
			v_etc := v_etc + coalesce(im_estimate_to_complete__project_user_etc(row.project_id, p_user_id), 0.0);
		END LOOP;
		return v_etc;
	END IF;

	-- Calculate the sum of all assignments to the project
	select	sum(bom.percentage) into v_assigned_percentage_task
	from	acs_rels r,
		im_biz_object_members bom
	where	r.rel_id = bom.rel_id and
		r.object_id_one = p_project_id and
		bom.percentage is not null;
--	RAISE NOTICE 'im_estimate_to_complete__project_user_etc: assigned_percentage_task=%', v_assigned_percentage_task;

	IF v_assigned_percentage_task = 0 THEN v_assigned_percentage_task := 0.01; END IF;
	FOR row IN
		select	pe.person_id,
			bom.percentage,
			v_planned_hours_task * (100 - v_percent_completed) * bom.percentage * 0.01 / v_assigned_percentage_task as planned_etc_user,
			(select	etc_units
			from	im_estimate_to_completes te
			where	te.etc_project_id = p.project_id and 
				te.etc_user_id = pe.person_id
			) as logged_etc_user
		from	acs_rels r,
			im_biz_object_members bom,
			persons pe,
			im_projects p
		where	p.project_id = p_project_id and
			pe.person_id = p_user_id and
			r.rel_id = bom.rel_id and
			r.object_id_one = p.project_id and
			r.object_id_two = pe.person_id and
			bom.percentage is not null
	LOOP
--		RAISE NOTICE 'im_estimate_to_complete__project_user_etc: %: ass=%, planned_etc=%, logged_etc=%', 
--			im_name_from_user_id(row.person_id), row.percentage, row.planned_etc_user, row.logged_etc_user;
		v_etc_user := row.planned_etc_user;
		IF row.logged_etc_user is not null THEN v_etc_user := row.logged_etc_user; END IF;

		v_etc := v_etc + v_etc_user;
	END LOOP;

	return v_etc;
end;$$ language 'plpgsql';

-- select im_estimate_to_complete__project_user_etc(55707, 55793); -- Arcus MM 2
-- select im_estimate_to_complete__project_user_etc(55724, 55793);


-- -------------------------------------------------------------------
-- Gantt TaskList
-- -------------------------------------------------------------------

--
-- Wide View in "Tasks" page, including Description
--
delete from im_view_columns where view_id = 910;
delete from im_views where view_id = 910;
insert into im_views (view_id, view_name, visible_for) values (910, 'im_timesheet_task_list', 'view_projects');


insert into im_view_columns (column_id, view_id, group_id, column_name, column_render_tcl,
extra_select, extra_where, sort_order, visible_for) values (91022,910,NULL, 
'"<input id=list_check_all type=checkbox name=_dummy>"',
'"<input type=checkbox name=task_id.$task_id id=tasks,$task_id>"', '', '', -1, '');

insert into im_view_columns (column_id, view_id, group_id, column_name, column_render_tcl,
extra_select, extra_where, sort_order, visible_for) values (91000,910,NULL,
'"<a href=$wiki_url/package-intranet-task-management#task_status>[im_gif help "Progress"]</a>"',
'[im_task_management_color_code_gif $progress_status_color_code]',
'im_task_management_color_code(t.task_id) as progress_status_color_code',
'',0,'im_package_exists_p "intranet-task-management"');

insert into im_view_columns (column_id, view_id, group_id, column_name, column_render_tcl,
extra_select, extra_where, sort_order, visible_for) values (91002,910,NULL,'"Task Name"',
'"<nobr>$indent_html$gif_html<a href=$object_url target=_blank>$task_name</a></nobr>"','','',20,'');

insert into im_view_columns (column_id, view_id, group_id, column_name, column_render_tcl,
extra_select, extra_where, sort_order, visible_for) values (91004,910,NULL,'Material',
'"<a href=/intranet-material/new?[export_vars -url {material_id return_url}] target=_blank>$material_nr</a>"',
'','',40,'set a 0');

insert into im_view_columns (column_id, view_id, group_id, column_name, column_render_tcl,
extra_select, extra_where, sort_order, visible_for) values (91006,910,NULL,'"CC"',
'"<a href=/intranet-cost/cost-centers/new?[export_vars -url {cost_center_id return_url}] target=_blank>$cost_center_code</a>"',
'','',60,'set a 0');

insert into im_view_columns (column_id, view_id, group_id, column_name, column_render_tcl,
extra_select, extra_where, sort_order, visible_for) values (91007,910,NULL,'"Start"',
'"<nobr>[string range $start_date 0 9]</nobr>"','','',80,'');

insert into im_view_columns (column_id, view_id, group_id, column_name, column_render_tcl,
extra_select, extra_where, sort_order, visible_for) values (91008,910,NULL,'"End"',
'"<nobr><font color=$end_date_color>[string range $end_date 0 9]</font></nobr>"',
'CASE WHEN child.end_date < now() and coalesce(child.percent_completed,0) < 100 THEN ''red'' ELSE ''black'' END as end_date_color','',100,'');

insert into im_view_columns (column_id, view_id, group_id, column_name, column_render_tcl,
extra_select, extra_where, sort_order, visible_for) values (91018,910,NULL,'Status',
'$status_select','','',120,'');

insert into im_view_columns (column_id, view_id, group_id, column_name, column_render_tcl,
extra_select, extra_where, sort_order, visible_for) values (91010,910,NULL,'Plan',
'$planned_hours_input','','',200,'');

insert into im_view_columns (column_id, view_id, group_id, column_name, column_render_tcl,
extra_select, extra_where, sort_order, visible_for) values (91012,910,NULL,'Bill',
'"<input type=textbox size=3 name=billable_units.$task_id value=$billable_units>"','','',220,'set a 0');

insert into im_view_columns (column_id, view_id, group_id, column_name, column_render_tcl,
extra_select, extra_where, sort_order, visible_for) values (91014,910,NULL,'Log',
'"[if {$planned_units > 0.0} { set t "
<div align=right><a href=[export_vars -base $timesheet_report_url {project_id {level_of_detail 5}}] target=_blank>
<font color=$log_color>$reported_units_cache / [expr round(100.0 * $reported_units_cache / $planned_units)]%</font></a></div>
" } else { set t "
<div align=right><a href=[export_vars -base $timesheet_report_url {project_id {level_of_detail 5}}] target=_blank>
<font color=$log_color>$reported_units_cache / -</font></a></div>
" }]"',
'CASE WHEN child.reported_hours_cache > child.percent_completed * t.planned_units / 100.0 
THEN ''red'' ELSE ''#235c96'' END as log_color','',240,'');

insert into im_view_columns (column_id, view_id, group_id, column_name, column_render_tcl,
extra_select, extra_where, sort_order, visible_for) values (91016,910,NULL,'UoM',
'$uom','','',260,'set a 0');

insert into im_view_columns (column_id, view_id, group_id, column_name, column_render_tcl,
extra_select, extra_where, sort_order, visible_for) values (91021,910,NULL, 'Done',
'$percent_done_input', '','',400,'');

insert into im_view_columns (column_id, view_id, group_id, column_name, column_render_tcl,
extra_select, extra_where, sort_order, visible_for) values (91023,910,NULL,'"ETC<br>Plan"',
'"<div align=right>[expr round((100.0 - $percent_completed) * $planned_units * 0.1) / 10.0]</div>"','',
'',500,'im_table_exists im_estimate_to_completes');

insert into im_view_columns (column_id, view_id, group_id, column_name, column_render_tcl,
extra_select, extra_where, sort_order, visible_for) values (91025,910,NULL,'"ETC<br>Earned V."',
'"<div align=right>$etc_eva</div>"','
CASE WHEN child.percent_completed > 0.0 THEN round((child.reported_hours_cache * (100.0 - child.percent_completed) / child.percent_completed)::numeric,1) ELSE 0 END as etc_eva',
'',510,'im_table_exists im_estimate_to_completes');

insert into im_view_columns (column_id, view_id, group_id, column_name, column_render_tcl,
extra_select, extra_where, sort_order, visible_for) values (91027,910,NULL,'"ETC<br>Manual"',
'"<div align=right>$etc_manual</div>"','
round(im_estimate_to_complete__user_etc(child.project_id),1) as etc_manual',
'',520,'im_table_exists im_estimate_to_completes');




-- insert into im_view_columns (column_id, view_id, group_id, column_name, column_render_tcl,
-- extra_select, extra_where, sort_order, visible_for) values (91020,910,NULL, 'Description', 
-- '[string_truncate -len 80 " $description"]', '','',300,'');




