# /packages/intranet-estimate-to-complete/tcl/intranet-estimate-to-complete-procs.tcl
#
# Copyright (C) 1998-2004 various parties

ad_library {
    Estimate to Complete

    @author frank.bergmann@project-open.com
}

# ---------------------------------------------------------------------
#
# ---------------------------------------------------------------------

ad_proc -public im_estimate_to_complete_task_details_component {
    task_id
} {
    Returns information about ETC and other estimation types.
} {
    db_1row task_info "
	select	p.*,
		t.*,
		im_estimate_to_complete__user_etc(p.project_id) as etc_entered,
		p.start_date::date as start_date_pretty,
		p.end_date::date as end_date_pretty,
		now()::date as today_pretty,
		extract(epoch from p.start_date) as start_date_epoch,
		extract(epoch from p.end_date) as end_date_epoch,
		extract(epoch from (p.end_date - p.start_date)) as duration_epoch,
		extract(epoch from now()) as today_epoch,
		(	select	sum(pbom.percentage)
			from	acs_rels pr,
				im_biz_object_members pbom
			where	pr.rel_id = pbom.rel_id and
				pr.object_id_one = p.project_id
		) as assigned_percentage
	from	im_projects p,
		im_timesheet_tasks t
	where	p.project_id = :task_id and
		p.project_id = t.task_id
    "

    if {"" eq $reported_hours_cache} { set reported_hours_cache 0.0 }
    if {"" eq $percent_completed} { set percent_completed 0.0 }

    if {"" eq $start_date_epoch || "" eq $duration_epoch} {
	return "
		<h2>Incomplete Task</h2>
		<p>start-date, end-date or percent-completed are empty, so we can't show reasonable ETC numbers.<br>
		Please correct the task setup.<br>
		</p>
        "
    }

    # -----------------------------------------------------
    # 
    set result [list]
    lappend result [list "Time Planning"]
    lappend result [list "Start" $start_date_pretty "Scheduled start of task"]
    lappend result [list "End" $end_date_pretty "Scheduled end of task"]

    lappend result [list "Time Passed" [expr ($today_epoch - $start_date_epoch) / (24 * 2600)] "Days since task start.<br>Negative values indicatate that the task will start in the future."]
    if {0 == $duration_epoch} { set duration_epoch [expr 24 * 3600] }
    set today_perc [expr 100.0 * ($today_epoch - $start_date_epoch) / $duration_epoch]
    lappend result [list "Time Passed %" $today_perc "Progress of the task as a percentage between start- and end date.<br>Values &lt; 0 indicate that the task will start in the future.<br>Values &gt; 100 indicate that end date of the task has passed."]

    lappend result [list "Hour Based Planning"]
    lappend result [list "Planned Hours" $planned_units "Estimation of the work needed to complete the task.<br>Typically imported from MS-Project or entered manually."]
    lappend result [list "Billable Hours" $billable_units "Normally same as planned units, but used for invoicing customers."]
    lappend result [list "Logged Hours" $reported_hours_cache "Hours logged by employees in the timesheet."]

    # Deal with division by zero
    set reported_hours_perc 0.0
    if {$reported_hours_cache > 0.0} {
	set reported_hours_perc "undefined"
	if {$planned_units > 0.0} {
	    set reported_hours_perc [expr 100.0 * $reported_hours_cache / $planned_units]
	}
    }

    lappend result [list "Logged Hours %" $reported_hours_perc "Logged hours as a percentage of planned hours.<br>This corresponds to the resources consumed<br>(ignoring resources other than logged hours)."]
    lappend result [list "Percent Completed" $percent_completed "How much of the task has been finished (percent)?<br>Entered manually by the PM normally.<br>Should be greater or equal to logged hours %, otherwise the task is late."]
    lappend result [list "Resources Assigned" $assigned_percentage "How many resources have been assigned to the task (percent)?<br>One resource counts as 100%."]

    # -----------------------------------------------------   
    #
    lappend result [list "Estimate to Complete"]
    if {"" eq $planned_units} { set planned_units 0.0 }
    set etc_planned [expr $planned_units * (100.0 - $percent_completed) / 100.0]
    lappend result [list "Estimate to Complete<br>(Planned)" $etc_planned "
	Based on the number of hours left to work, according to planned hours and percent completed.<br>
	etc_planned = planned_hours * (100 - percent_completed) / 100"]
    set etc_eva $planned_units
    if {$percent_completed > 0.0} {
	set etc_eva [expr (100.0 - $percent_completed) * $reported_hours_cache / $percent_completed]
    }
    lappend result [list "Estimate to Complete<br>(Earned Value)" $etc_eva "
	Based on the logged hours that it took to achieve the current percent completed.<br>
	etc_eva = (100 - percent_completed) * logged_hours / percent_completed"]
    lappend result [list "Estimate to Complete<br>(Manually Entered)" $etc_entered "
        Based on manually entered ETC values (during timesheet entry), taking planned ETC as a default if no estimates are available.<br>
        Watch out for outdated ETC estimates!"]


    # -----------------------------------------------------   
    #
    lappend result [list "Estimate at Completion"]
    lappend result [list "Estimate at Completion<br>(Planned)" $planned_units "= Planned Hours."]
    set eac_eva [expr $reported_hours_cache + $etc_eva]
    lappend result [list "Estimate at Completion<br>(Earned Value)" $eac_eva "Based on the logged hours that it took to achieve the current percent completed."]
    set eac_entered ""
    if {"" ne $etc_entered} {
	set eac_entered [expr $reported_hours_cache + $etc_entered]
    }
    lappend result [list "Estimate at Completion<br>(Manually Entered)" $eac_entered "Based on manually entered ETC values (during timesheet entry), taking planned ETC as a default if no estimates are available.<br>Watch out for outdated ETC estimates!"]

    set html ""
    foreach tuple $result {
        set key [lindex $tuple 0]
	regsub -all {<[^<>]+>} $key "_" key_underscored
	regsub -all {[\%\(\)]} $key_underscored "_" key_underscored
	regsub -all { } $key_underscored "_" key_underscored
	regsub -all {__} $key_underscored "_" key_underscored
	regsub -all {_$} $key_underscored "" key_underscored
        set key_l10n [lang::message::lookup "" intranet-timesheet2-tasks.$key_underscored $key]

	if {[llength $tuple] eq 1} {
	    append html "<tr valign=top><td colspan=3 align=center><b>$key_l10n</b></td></tr>\n"
	    continue
	}

	set value [lindex $tuple 1]
	if {"" ne $value && [string is double $value]} {
	    set value [expr round($value * 10.0) / 10.0]
	}
        append html "<tr valign=top><td><nobr>$key_l10n</nobr></td><td align=right><nobr>$value</nobr></td><td>[lindex $tuple 2]</tr>\n"
    }

    return "
    <table cellspacing=2 cellpadding=2>
    <th><tr class=rowtitle>
    <td class=rowtitle>[lang::message::lookup "" intranet-timesheet2.Indicator "Indicator"]</td>
    <td class=rowtitle>[lang::message::lookup "" intranet-timesheet2.Value "Value"]</td>
    <td class=rowtitle>[lang::message::lookup "" intranet-timesheet2.Comment "Comment"]</td>
    </tr></th>
    $html</table>
    "

}


ad_proc -public im_estimate_to_complete_task_history_component {
    task_id
} {
    Returns information about ETC and other estimation types.
} {
    set current_user_id [auth::require_login]

    set history_sql "
select
	creation_date::date as day,
	round(task_planned_hours_audit,1) as task_planned_hours,
	round(task_completed_percent_audit,1) as task_percent_completed,
	im_name_from_user_id(etc_user_id) as user_name,
	round(task_planned_hours_audit * user_assigned_percent_audit / task_assigned_percent_audit,1) as user_planned_hours,
	round(100.0 * user_assigned_percent_audit / task_assigned_percent_audit,1) as percent_hours_assigned_to_user,
	round(logged_hours, 1) as user_logged_hours,
	round(etc_units,1) as estimate_to_complete_manual,
	round(logged_hours + etc_units,1) as estimate_at_completion_based_on_etc,
	round(100.0 * 
		(logged_hours + etc_units - (task_planned_hours_audit * user_assigned_percent_audit / task_assigned_percent_audit)) 
		/ greatest(task_planned_hours_audit * user_assigned_percent_audit / task_assigned_percent_audit, 0.0000001)
	,1) as eac_deviation_percent
from	(select	e.*,
		o.creation_date,
		(	select	round(sum(h.hours),1)
			from	im_hours h
			where	h.project_id = :task_id and
				h.user_id = :current_user_id
		) as logged_hours,
		(select	sum(percentage)
		from	acs_rels r,
			im_biz_object_members bom
		where	r.rel_id = bom.rel_id and
			r.object_id_two = :current_user_id and
			r.object_id_one = :task_id
		) as user_percent_assigned,
		(select	greatest(sum(percentage), 0.01)
		from	acs_rels r,
			im_biz_object_members bom
		where	r.rel_id = bom.rel_id and
			r.object_id_one = :task_id
		) as total_percent_assigned
	from	acs_objects o,
		im_estimate_to_completes e
	where	o.object_id = e.etc_id and
		e.etc_project_id = :task_id
	) t
order by
	im_name_from_user_id(etc_user_id),
	creation_date
"

    return [im_ad_hoc_query -format html $history_sql]
}

